// Functions for fitness computation and for recombination:

#include "depression.h"
#include <vector>
#include <algorithm>
#include <cmath>
using namespace std;

extern MTRand rnd;


//Computes the fitness of a diploid with chromosomes c1 and c2:

double fitness(chr &c1, chr &c2, double wHe, double wHo, double wAA, double wAD, double wDD)
{
	int s1 = c1.sel.size() - 1;
	int s2 = c2.sel.size() - 1;
	
	int nHo = 0;
	int nHe1 = 0;
	int nHe2 = 0;
	int nHet;
	
	double HeHe, HeHo, HoHo;
	
	while ((s1 > -1) || (s2 > -1))
	{
		if (s1 == -1)
		{
			nHe2 += s2 + 1;
			break;
		}
		
		if (s2 == -1)
		{
			nHe1 += s1 + 1;
			break;
		}
		
		if (c1.sel[s1] > c2.sel[s2])
		{
			nHe1++;
			s1--;
			continue;
		}
		
		if (c1.sel[s1] < c2.sel[s2])
		{
			nHe2++;
			s2--;
			continue;
		}
		
		nHo++;
		s1--;
		s2--;
	}
	
	c1.nHe = nHe1;
	c2.nHe = nHe2;
	nHet = nHe1 + nHe2;
	
	HeHe = nHet * (nHet - 1) / 2;
	HeHo = nHet * nHo;
	HoHo = nHo * (nHo - 1) / 2;
	
	return pow(wHe, nHet) * pow(wHo, nHo) * pow (wAA, HeHe + 2*HeHo + 4*HoHo) * pow(wAD, HeHo + 4*HoHo) * pow(wDD, HoHo);
}



// Constructs chromosome "res" by recombining chromosomes c1 and c2
// "Sz" is the map length L (average number of cross-overs per meiosis).
// The modifier gene is inherited from chromosome c1.
// If the number of mutations in the parent is less than L, the recombining
// chromosome is created by walking along the genome and calculating the probability
// that a cross-over occurs between each pair of adjacent mutations (strategy 1).
// If the number of mutations is greater than L, the number and position of cross-overs 
// are sampled randomly (strategy 2).

void rec(chr &res, chr &c1, chr &c2, double Sz)
{
	int sz1 = c1.sel.size();
	int sz2 = c2.sel.size();
	int s1 = sz1 - 1;
	int s2 = sz2 - 1;
	
	res.sel.clear();
	res.mod = c1.mod;

	// strategy 1 (see above):
	
	if (sz1 + sz2 < Sz)
	{
		bool str = 0;
		double newend;
	
		double End = Sz; // position of modifier locus (=L). Note that positions of 
				 // mutations are between 0 and 2L, so the modifier is located
				 // at the mid-point of the chromosome. This could be changed here.
	
		// one starts with mutations on the left side of the modifier:

		if (sz1 > 0)
		{
			while ((s1 > -1) && (c1.sel[s1] > End))
				s1--;
		}
		if (sz2 > 0)
		{
			while ((s2 > -1) && (c2.sel[s2] > End))
				s2--;
		}

		// c1.sel[s1] and c2.sel[s2] are now the mutations nearest to the modifier on the left
		// sides of chromosomes c1 and c2. The mutations on the offspring chromosome are generated
                // by walking along the chromosome from position L to position 0 (from right to left),
		// switching between parental chromosomes when a recombination event occurs.
		
		while ((s1 > -1) || (s2 > -1))
		{
	
			// if no more mutations on chromosome 1:

			if (s1 == -1)
			{
				newend = c2.sel[s2]; // position of nearest mutation on chromosome 2
			
				if (str == 0) // if we are on chromosome 1
				{
					// if a recombination occurs between current position 
					// and the position of nearest mutation:

					if (rnd.rand() < (1 - exp(newend - End)) / 2)
					{
						str = 1;  // we move to chromosome 2
						res.sel.push_back(newend); // the mutation is incorporated in offspring
					}
				}
				else // if we are on chromosome 2
				{
					if (rnd.rand() < (1 - exp(newend - End)) / 2)
						str = 0; // if recombination, we move to chromosome 1
					else
						res.sel.push_back(newend); // the mutation is incorporated
				}
				s2--;
				End = newend;
				continue;
			}
	
		
			// if no more mutations on chromosome 2: same as above

			else if (s2 == -1)
			{
				newend = c1.sel[s1];
			
				if (str == 0)
				{
					if (rnd.rand() < (1 - exp(newend - End)) / 2)
						str = 1;
					else
						res.sel.push_back(newend);
				}
				else
				{
					if (rnd.rand() < (1 - exp(newend - End)) / 2)
					{
						str = 0;
						res.sel.push_back(newend);
					}
				}
				s1--;
				End = newend;
				continue;
			}
		
			// if the nearest mutation is on chromosome 2:

			else if (c1.sel[s1] < c2.sel[s2])
			{
				newend = c2.sel[s2]; // position of nearest mutation
			
				// if we are on chromosome 1

				if (str == 0)
				{
					// if a recombination occurs between current position 
					// and the position of the mutation:

					if (rnd.rand() < (1 - exp(newend - End)) / 2)
					{
						str = 1;
						res.sel.push_back(newend);
					}
				}
				else
				{
					if (rnd.rand() < (1 - exp(newend - End)) / 2)
						str = 0;
					else
						res.sel.push_back(newend);
				}
				s2--;
				End = newend;
				continue;
			}

			// if the nearest mutation is on chromosome 1:
		
			else if (c1.sel[s1] > c2.sel[s2])
			{
				newend = c1.sel[s1];
			
				if (str == 0)
				{
					if (rnd.rand() < (1 - exp(newend - End)) / 2)
						str = 1;
					else
						res.sel.push_back(newend);
				}
				else
				{
					if (rnd.rand() < (1 - exp(newend - End)) / 2)
					{
						str = 0;
						res.sel.push_back(newend);
					}
				}
				s1--;
				End = newend;
				continue;
			}

			// if the nearest mutation is homozygous:
		
			newend = c1.sel[s1];
			res.sel.push_back(newend);
			if (str == 0)
				if (rnd.rand() < (1 - exp(newend - End)) / 2)
					str = 1;
			else
				if (rnd.rand() < (1 - exp(newend - End)) / 2)
					str = 0;
			s1--;
			s2--;
			End = newend;
		}

		// we now deal with mutations on the right side of the modifier locus:
	
		End = Sz; // modifier at the mid-point of the chromosome (position L)
		s1 = 0;
		s2 = 0;
		str = 0;
	
		if (sz1 > 0)
		{
			while ((s1 < sz1) && (c1.sel[s1] < End))
				s1++;
		}
		if (sz2 > 0)
		{
			while ((s2 < sz2) && (c2.sel[s2] < End))
				s2++;
		}

		// c1.sel[s1] and c2.sel[s2] are now the mutations nearest to the modifier on the right
		// sides of chromosomes c1 and c2. The mutations on the offspring chromosome are generated
		// by walking along the chromosome from position L to position 2L (from left to right),
		// switching between parental chromosomes when a recombination event occurs.
	
		while ((s1 < sz1) || (s2 < sz2))
		{
	
			// if no more mutation on chromosome 1:

			if (s1 == sz1)
			{
				newend = c2.sel[s2];
			
				if (str == 0)
				{
					if (rnd.rand() < (1 - exp(End - newend)) / 2)
					{
						str = 1;
						res.sel.push_back(newend);
					}
				}
				else
				{
					if (rnd.rand() < (1 - exp(End - newend)) / 2)
						str = 0;
					else
						res.sel.push_back(newend);
				}
				s2++;
				End = newend;
				continue;
			}
	
		
			// if no more mutation on chromosome 2:

			else if (s2 == sz2)
			{
				newend = c1.sel[s1];
			
				if (str == 0)
				{
					if (rnd.rand() < (1 - exp(End - newend)) / 2)
						str = 1;
					else
						res.sel.push_back(newend);
				}
				else
				{
					if (rnd.rand() < (1 - exp(End - newend)) / 2)
					{
						str = 0;
						res.sel.push_back(newend);
					}
				}
				s1++;
				End = newend;
				continue;
			}
		
			// if nearest mutation is on chromosome 2:

			else if (c1.sel[s1] > c2.sel[s2])
			{
				newend = c2.sel[s2];
			
				if (str == 0)
				{
					if (rnd.rand() < (1 - exp(End - newend)) / 2)
					{
						str = 1;
						res.sel.push_back(newend);
					}
				}
				else
				{
					if (rnd.rand() < (1 - exp(End - newend)) / 2)
						str = 0;
					else
						res.sel.push_back(newend);
				}
				s2++;
				End = newend;
				continue;
			}
		
			// if nearest mutation is on chromosome 1:

			else if (c1.sel[s1] < c2.sel[s2])
			{
				newend = c1.sel[s1];
			
				if (str == 0)
				{
					if (rnd.rand() < (1 - exp(End - newend)) / 2)
						str = 1;
					else
						res.sel.push_back(newend);
				}
				else
				{
					if (rnd.rand() < (1 - exp(End - newend)) / 2)
					{
						str = 0;
						res.sel.push_back(newend);
					}
				}
				s1++;
				End = newend;
				continue;
			}
		
			// if nearest mutation is homozygous:

			newend = c1.sel[s1];
			res.sel.push_back(newend);
			if (str == 0)
				if (rnd.rand() < (1 - exp(End - newend)) / 2)
					str = 1;
			else
				if (rnd.rand() < (1 - exp(End - newend)) / 2)
					str = 0;
			s1++;
			s2++;
			End = newend;
		}
	}
	
	// strategy 2 (see above):

	else
	{
		int j;
		chr Co;
	
		double twoS = 2 * Sz;
	
		// number of cross-overs:
	
		int nbCo = int(poisdev(Sz));

		// positions of cross-overs (between 0 and 2L) are put in the vector Co.sel
		
		for (j = 0; j < nbCo; j++)
			Co.sel.push_back(twoS * rnd.rand());
		sort(Co.sel.begin(), Co.sel.end());
	
		// Counting the number of cross-overs on the right of the modifier locus:

		int cmpt = 0;
		j = Co.sel.size() - 1;
		while ((j > -1) && (Co.sel[j] > Sz))
		{
			cmpt++;
			j--;
		}
	
		// if the number of cross-overs on the right is even:

		if (cmpt % 2 == 0)
		{	
			// for each cross-over (starting on extreme right):

			for (j = 1; j <= cmpt; j++)
			{
				if (j % 2 == 1)
				{
					// all mutations on the right of cross-over on chromosome 1 are incorporated

					while ((s1 > -1) && (c1.sel[s1] > Co.sel[nbCo - j]))
					{
						res.sel.push_back(c1.sel[s1]);
						s1--;
					}
					while ((s2 > -1) && (c2.sel[s2] > Co.sel[nbCo - j]))
						s2--;
				}
				else
				{
					// all mutations on the right of cross-over on chromosome 2 are incorporated

					while ((s2 > -1) && (c2.sel[s2] > Co.sel[nbCo - j]))
					{
						res.sel.push_back(c2.sel[s2]);
						s2--;
					}
					while ((s1 > -1) && (c1.sel[s1] > Co.sel[nbCo - j]))
						s1--;
				}
			}
		}
	
		// if the number of cross-overs on the right is odd:

		else
		{	
			// for each cross-over (starting on extreme right):

			for (j = 1; j <= cmpt; j++)
			{
				if (j % 2 == 1)
				{
					// all mutations on the right of cross-over on chromosome 2 are incorporated

					while ((s2 > -1) && (c2.sel[s2] > Co.sel[nbCo - j]))
					{
						res.sel.push_back(c2.sel[s2]);
						s2--;
					}
					while ((s1 > -1) && (c1.sel[s1] > Co.sel[nbCo - j]))
						s1--;
				}
				else
				{
					// all mutations on the right of cross-over on chromosome 1 are incorporated

					while ((s1 > -1) && (c1.sel[s1] > Co.sel[nbCo - j]))
					{
						res.sel.push_back(c1.sel[s1]);
						s1--;
					}
					while ((s2 > -1) && (c2.sel[s2] > Co.sel[nbCo - j]))
						s2--;
				}
			}
		}
	
		// mutations between the modifier and the nearest cross-over on the right (on chromosome 1):

		while ((s1 > -1) && (c1.sel[s1] > Sz))
		{
			res.sel.push_back(c1.sel[s1]);
			s1--;
		}
		while ((s2 > -1) && (c2.sel[s2] > Sz))
			s2--;

		// number of cross-overs on the left of the modifier locus:
		
		int frst = nbCo - cmpt;

		// for each cross-over (starting with the nearest to the modifier on the left):
		
		for (j = 1; j <= frst; j++)
		{
			if (j % 2 == 1)
			{
				// all mutations on the right of cross-over on chromosome 1 are incorporated

				while ((s1 > -1) && (c1.sel[s1] > Co.sel[frst - j]))
				{
					res.sel.push_back(c1.sel[s1]);
					s1--;
				}
				while ((s2 > -1) && (c2.sel[s2] > Co.sel[frst - j]))
					s2--;
			}
			else
			{
				// all mutations on the right of cross-over on chromosome 2 are incorporated

				while ((s2 > -1) && (c2.sel[s2] > Co.sel[frst - j]))
				{
					res.sel.push_back(c2.sel[s2]);
					s2--;
				}
				while ((s1 > -1) && (c1.sel[s1] > Co.sel[frst - j]))
					s1--;
			}
		}

		// mutations on the left of the left-most cross-over:
	
		if (frst % 2 == 0)
			while (s1 > -1)
			{
				res.sel.push_back(c1.sel[s1]);
				s1--;
			}
		else
			while (s2 > -1)
			{
				res.sel.push_back(c2.sel[s2]);
				s2--;
			}

	}

	// sorts mutations on offspring chromosome:
	
	sort(res.sel.begin(), res.sel.end());
}
