#include "depression.h"
#include <iostream>
#include <fstream>
#include <sstream>
#include <ctime>
#include <vector>
#include <algorithm>
#include <cmath>
using namespace std;

extern MTRand rnd;
extern FILE * fichierE;
extern FILE * fichierS;

/* function recursion: iterates the life cycle. The selfing rate is kept constant 
 during the first NbPrelimv generations, and is free to evolve during the next
 NbGenv generations.
 Other parameters are:
 	Nv: population size
	av: initial selfing rate
	rhov: rate of pollen discounting
	sv: selection coefficient of deleterious alleles
    hv: dominance coefficient of deleterious alleles
    eaav: additive-by-additive epistasis
    eadv: addiive-by-dominance epistasis
    eddv: dominance-by-dominance epistasis
	Uv: deleterious mutation rate per haploid genome
	Lv: genome map length (average number of cross-overs at meiosis)
	mutMv: mutation rate at modifier locus
	mutMstepv: step size for modifier mutation
	pasprv: number of generations between measures during preliminary generations
	pasv: number of generations between measures after preliminary generations
	sampleSv: sample size for estimating inbreeding depression
*/


void recursion(int Nv, double av, double rhov, double sv, double hv, double eaav, double eadv, double eddv, 
		double Uv, double Lv, double mutMv, double mutMstepv, int NbPrelimv, int NbGenv, int pasprv, 
		int pasv, int sampleSv, int nov)
{
	int i, j, k, gen, mut, nbMig, chr1, chr2, ind, nb, nb2, nb3, nb4, prem1, prem2, par1;
	double cmpt1, w, wbar, wmax, wmaxm, le, rd, rd2;
	double wout, wself, woutsel, wselfsel, slf, mutPos;
	chr off1, off2;

	int pas_c = 50;
	int nbFix = 0;
	
	int twoN = 2*Nv;
	double twoL  = 2*Lv;
	double Whet = 1 - hv * sv;
	double Whom = 1 - sv;
	double Waa = 1 + eaav;
	double Wad = 1 + eadv;
	double Wdd = 1 + eddv;
	
	// population: table of 2N chromosomes:

	chr * pop = new chr [twoN];
	chr * temp = new chr [twoN];

	// table of fitness values (for all individuals)
	
	double * Wij = new double [Nv];
	double * Wijm = new double [Nv];

	// creating name of output file (which indicates the parameter values):

	char nomFichier[256];
	stringstream nomF;
	nomF << "result_N" << Nv << "_a" << av << "_rho" << rhov << "_s" << sv << "_h" << hv
		<< "_eaa" << eaav << "_ead" << eadv << "_edd" << eddv
		 << "_U" << Uv << "_L" << Lv << "_" << nov << ".txt";
	nomF >> nomFichier; 
	ofstream fout(nomFichier);

	// keeping track of starting time:
	
	time_t debut, fin;
	struct tm *ptr;
	debut = time(0);
	
	//initialization: all modifier alleles code for selfing rate av:
	
	for (i = 0; i < twoN; i++)
		pop[i].mod = av;
	
	//preliminary generations: no mutation at modifier locus
	
	for (gen = 0; gen < NbPrelimv; gen++)
	{ 
		// removing fixed mutations:

		if (gen % pas_c == 0)
			for (i = pop[0].sel.size()-1; i >= 0 ; i--)
			{
				mutPos = pop[0].sel[i];
				j = 1;
				while ((j < twoN) && 
					  (find(pop[j].sel.begin(), pop[j].sel.end(), mutPos) != pop[j].sel.end())) 
					j++;
				if (j == twoN)
				{
					for (k = 0; k < twoN; k++)
						pop[k].sel.erase(find(pop[k].sel.begin(), pop[k].sel.end(), mutPos));
					nbFix++;
				}
			}

		//deleterious mutation:
		
		for (i = 0; i < twoN; i++) // for each chromosome
		{
			mut = poisdev(Uv); // number of new mutations on chromosome
			for (j = 0; j < mut; j++)
			{
				// each mutation has a random position between 0 and 2L:

				pop[i].sel.push_back(twoL * rnd.rand()); 
				sort(pop[i].sel.begin(), pop[i].sel.end());
			}
		}
		
		//fills table Wij with fitnesses of all individuals, and calculates average fitness (wbar):
		
		wbar = 0;
		wmax = 0;

		// for each individual j:
		for (j = 0; j < Nv; j++)
		{
			nb2 = 2*j;
			w = fitness(pop[nb2], pop[nb2 + 1], Whet, Whom, Waa, Wad, Wdd);
			wbar += w;
			Wij[j] = w;
				
			if (wmax < w)
				wmax = w;
		}
		wbar /= Nv;
		
		// sampling the next generation:
		
		for (ind = 0; ind < Nv; ind++)
		{
			nb4 = 2*ind;
			
			// sampling the mother:
				
			do{
				chr1 = int(rnd.randExc(twoN));
				prem1 = chr1/2;
			} while (rnd.rand() > Wij[prem1] / wmax);
				
			// recombination:
				
			if (chr1 % 2 == 0)
				rec(temp[nb4], pop[chr1], pop[chr1 + 1], Lv);
			else
				rec(temp[nb4], pop[chr1], pop[chr1 - 1], Lv);
				
			// sampling the father:

			rd = rnd.rand();
			
			//selfing:
			
			if (rd < av)
			{
				rd2 = rnd.rand();
				par1 = 2*prem1;
				
				if (rd2 < 0.5)
					rec(temp[nb4 + 1], pop[par1], pop[par1 + 1], Lv);
				else
					rec(temp[nb4 + 1], pop[par1 + 1], pop[par1], Lv);
			}

			//outcrossing:
				
			else
			{
				do{
					chr1 = int(rnd.randExc(twoN));
					prem1 = chr1/2;
				} while (rnd.rand() > Wij[prem1] / wmax);
				
				// recombination:
				
				if (chr1 % 2 == 0)
					rec(temp[nb4 + 1], pop[chr1], pop[chr1 + 1], Lv);
				else
					rec(temp[nb4 + 1], pop[chr1], pop[chr1 - 1], Lv);
			}
		}
			
		// measuring inbreeding depression every "pasprv" generations:
		
		if (gen % pasprv == 0)
		{
			// number of mutations per chromosome:
			
			le = 0;
			for (i = 0; i < twoN; i++)
				le += pop[i].sel.size();
				
			// measuring the average fitness of "sampleSv" selfed offspring,
			// and "sampleSv" offspring produced by outcrossing:
				
			wout = 0;
			wself = 0;
				
			// average fitness of "sampleS" selfed offspring:

			for (j = 0; j < sampleSv; j++)
			{
				chr1 = int(rnd.randExc(twoN));
				if (chr1 % 2 == 0)
					nb3 = chr1 + int(rnd.randExc(2));
				else
					nb3 = chr1 - int(rnd.randExc(2));
				if (chr1 % 2 == 0)
					rec(off1, pop[chr1], pop[chr1 + 1], Lv);
				else
					rec(off1, pop[chr1], pop[chr1 - 1], Lv);
				if (nb3 % 2 == 0)
					rec(off2, pop[nb3], pop[nb3 + 1], Lv);
				else
					rec(off2, pop[nb3], pop[nb3 - 1], Lv);
				wself += fitness(off1, off2, Whet, Whom, Waa, Wad, Wdd);
			}
				
			// average fitness of "sampleS" outcrossed offspring:

			for (j = 0; j < sampleSv; j++)
			{
				chr1 = int(rnd.randExc(twoN));
				chr2 = int(rnd.randExc(twoN));
				if (chr1 % 2 == 0)
					rec(off1, pop[chr1], pop[chr1 + 1], Lv);
				else
					rec(off1, pop[chr1], pop[chr1 - 1], Lv);
				if (chr2 % 2 == 0)
					rec(off2, pop[chr2], pop[chr2 + 1], Lv);
				else
					rec(off2, pop[chr2], pop[chr2 - 1], Lv);
				wout += fitness(off1, off2, Whet, Whom, Waa, Wad, Wdd);
			}
            
            woutsel = 0;
            wselfsel = 0;
            
            // inbreeding depression after selection (parents contribute in proportion of their fitness):
            // average fitness of "sampleS" selfed offspring:
            
            for (j = 0; j < sampleSv; j++)
            {
                do{
                    chr1 = int(rnd.randExc(twoN));
                    prem1 = chr1/2;
                } while (rnd.rand() > Wij[prem1] / wmax);
                if (chr1 % 2 == 0)
                    nb3 = chr1 + int(rnd.randExc(2));
                else
                    nb3 = chr1 - int(rnd.randExc(2));
                if (chr1 % 2 == 0)
                    rec(off1, pop[chr1], pop[chr1 + 1], Lv);
                else
                    rec(off1, pop[chr1], pop[chr1 - 1], Lv);
                if (nb3 % 2 == 0)
                    rec(off2, pop[nb3], pop[nb3 + 1], Lv);
                else
                    rec(off2, pop[nb3], pop[nb3 - 1], Lv);
                wselfsel += fitness(off1, off2, Whet, Whom, Waa, Wad, Wdd);
            }
            
            // average fitness of "sampleS" outcrossed offspring:
            
            for (j = 0; j < sampleSv; j++)
            {
                do{
                    chr1 = int(rnd.randExc(twoN));
                    prem1 = chr1/2;
                } while (rnd.rand() > Wij[prem1] / wmax);
                do{
                    chr2 = int(rnd.randExc(twoN));
                    prem2 = chr2/2;
                } while (rnd.rand() > Wij[prem2] / wmax);
                if (chr1 % 2 == 0)
                    rec(off1, pop[chr1], pop[chr1 + 1], Lv);
                else
                    rec(off1, pop[chr1], pop[chr1 - 1], Lv);
                if (chr2 % 2 == 0)
                    rec(off2, pop[chr2], pop[chr2 + 1], Lv);
                else
                    rec(off2, pop[chr2], pop[chr2 - 1], Lv);
                woutsel += fitness(off1, off2, Whet, Whom, Waa, Wad, Wdd);
            }
			
			// Writes selfing rate, mean fitness, mean nb of mutations per chromosome,
			// inbreeding depression and nb of fixed mutations in output file:

			fout << av << " " << wbar << " " << le / twoN
				 << " " << 1 - (wself / wout) << " " << 1 - (wselfsel / woutsel) << " " << nbFix << endl;
		}
        
        for (i = 0; i < twoN; i++)
            pop[i] = temp[i];
	}
	
	// letting selfing evolve:
	
	for (gen = 0; gen < NbGenv; gen++)
    {
        // eliminating fixed mutations:

        if (gen % pas_c == 0)
            for (i = pop[0].sel.size()-1; i >= 0 ; i--)
            {
                mutPos = pop[0].sel[i];
                j = 1;
                while ((j < twoN) &&
                    (find(pop[j].sel.begin(), pop[j].sel.end(), mutPos) != pop[j].sel.end()))
                    j++;
                if (j == twoN)
                {
                    for (k = 0; k < twoN; k++)
                        pop[k].sel.erase(find(pop[k].sel.begin(), pop[k].sel.end(), mutPos));
                    nbFix++;
                }
            }

        //mutation:
			
        for (i = 0; i < twoN; i++)
        {
            //modifier locus:
            if (rnd.rand() < mutMv)
            {
                pop[i].mod += mutMstepv * gasdev();
                if (pop[i].mod < 0)
                    pop[i].mod = 0;
                else if (pop[i].mod > 1)
                    pop[i].mod = 1;
            }
			
            //other loci:
            mut = int(poisdev(Uv));
            for (j = 0; j < mut; j++)
            {
                pop[i].sel.push_back(twoL * rnd.rand());
                sort(pop[i].sel.begin(), pop[i].sel.end());
            }
        }
			
        //fitnesses:
		
        wbar = 0;
        wmax = 0;
        wmaxm = 0;
			
        // for each individual j:
        for (j = 0; j < Nv; j++)
        {
            nb2 = 2*j;
            w = fitness(pop[nb2], pop[nb2 + 1], Whet, Whom, Waa, Wad, Wdd);
            slf = (pop[nb2].mod + pop[nb2 + 1].mod) / 2.0;  // selfing rate
            wbar += w;
            Wij[j] = w;
            Wijm[j] = w * (1 - slf * rhov);
				
            if (wmax < w)
                wmax = w;
            if (wmaxm < Wijm[j])
                wmaxm = Wijm[j];
        }
        wbar /= Nv;
			
        // sampling the next generation:
			
        for (ind = 0; ind < Nv; ind++)
        {
            nb4 = 2*ind;
				
            // selection of a mother:
					
            do{
                chr1 = int(rnd.randExc(twoN));
                prem1 = chr1/2;
            } while (rnd.rand() > Wij[prem1] / wmax);
					
            // recombination:
					
            if (chr1 % 2 == 0)
                rec(temp[nb4], pop[chr1], pop[chr1 + 1], Lv);
            else
                rec(temp[nb4], pop[chr1], pop[chr1 - 1], Lv);
					
            // selection of a parental chromosome (paternal gamete):
			
            rd = rnd.rand();
			
            // selfing:

            par1 = 2*prem1;
					
            if (rd < (pop[par1].mod + pop[par1 + 1].mod) / 2)
            {
                rd2 = rnd.rand();
						
                if (rd2 < 0.5)
                    rec(temp[nb4 + 1], pop[par1], pop[par1 + 1], Lv);
                else
                    rec(temp[nb4 + 1], pop[par1 + 1], pop[par1], Lv);
            }
			
            // outcrossing:
			
            else
            {
                do{
                    chr1 = int(rnd.randExc(twoN));
                    prem1 = chr1/2;
                } while (rnd.rand() > Wijm[prem1] / wmaxm);
				
                // recombination:
					
                if (chr1 % 2 == 0)
                    rec(temp[nb4 + 1], pop[chr1], pop[chr1 + 1], Lv);
                else
                    rec(temp[nb4 + 1], pop[chr1], pop[chr1 - 1], Lv);
            }
        }
			
        // measuring mean selfing rate, inbreeding depression and heterosis every "pasv" generations:
			
        if (gen % pasv == 0)
        {
            // number of mutations per chromosome, and selfing rate:
				
            le = 0;
            slf = 0;
            for (i = 0; i < twoN; i++)
            {
                le += pop[i].sel.size();
                slf += pop[i].mod;
            }
					
            // measuring the average fitness of "sampleSv" selfed offspring,
            // and "sampleSv" offspring produced by outcrossing:
				
            wout = 0;
            wself = 0;
				
            // average fitness of "sampleS" selfed offspring:
				
            for (j = 0; j < sampleSv; j++)
            {
                chr1 = int(rnd.randExc(twoN));
                if (chr1 % 2 == 0)
                    nb3 = chr1 + int(rnd.randExc(2));
                else
                    nb3 = chr1 - int(rnd.randExc(2));
                if (chr1 % 2 == 0)
                    rec(off1, pop[chr1], pop[chr1 + 1], Lv);
                else
                    rec(off1, pop[chr1], pop[chr1 - 1], Lv);
                if (nb3 % 2 == 0)
                    rec(off2, pop[nb3], pop[nb3 + 1], Lv);
                else
                    rec(off2, pop[nb3], pop[nb3 - 1], Lv);
                wself += fitness(off1, off2, Whet, Whom, Waa, Wad, Wdd);
            }
				
            // average fitness of "sampleS" outcrossed offspring:
				
            for (j = 0; j < sampleSv; j++)
            {
                chr1 = int(rnd.randExc(twoN));
                chr2 = int(rnd.randExc(twoN));
                if (chr1 % 2 == 0)
                    rec(off1, pop[chr1], pop[chr1 + 1], Lv);
                else
                    rec(off1, pop[chr1], pop[chr1 - 1], Lv);
                if (chr2 % 2 == 0)
                    rec(off2, pop[chr2], pop[chr2 + 1], Lv);
                else
                    rec(off2, pop[chr2], pop[chr2 - 1], Lv);
                wout += fitness(off1, off2, Whet, Whom, Waa, Wad, Wdd);
            }
            
            woutsel = 0;
            wselfsel = 0;
            
            // inbreeding depression measured after selection:
            // average fitness of "sampleS" selfed offspring:
            
            for (j = 0; j < sampleSv; j++)
            {
                do{
                    chr1 = int(rnd.randExc(twoN));
                    prem1 = chr1/2;
                } while (rnd.rand() > Wij[prem1] / wmax);
                    if (chr1 % 2 == 0)
                        nb3 = chr1 + int(rnd.randExc(2));
                    else
                        nb3 = chr1 - int(rnd.randExc(2));
                if (chr1 % 2 == 0)
                    rec(off1, pop[chr1], pop[chr1 + 1], Lv);
                else
                    rec(off1, pop[chr1], pop[chr1 - 1], Lv);
                if (nb3 % 2 == 0)
                    rec(off2, pop[nb3], pop[nb3 + 1], Lv);
                else
                    rec(off2, pop[nb3], pop[nb3 - 1], Lv);
                wselfsel += fitness(off1, off2, Whet, Whom, Waa, Wad, Wdd);
            }
            
            // average fitness of "sampleS" outcrossed offspring:
            
            for (j = 0; j < sampleSv; j++)
            {
                do{
                    chr1 = int(rnd.randExc(twoN));
                    prem1 = chr1/2;
                } while (rnd.rand() > Wij[prem1] / wmax);
                    do{
                        chr2 = int(rnd.randExc(twoN));
                        prem2 = chr2/2;
                    } while (rnd.rand() > Wij[prem2] / wmax);
                        if (chr1 % 2 == 0)
                            rec(off1, pop[chr1], pop[chr1 + 1], Lv);
                        else
                            rec(off1, pop[chr1], pop[chr1 - 1], Lv);
                if (chr2 % 2 == 0)
                    rec(off2, pop[chr2], pop[chr2 + 1], Lv);
                else
                    rec(off2, pop[chr2], pop[chr2 - 1], Lv);
                woutsel += fitness(off1, off2, Whet, Whom, Waa, Wad, Wdd);
            }
				
            // Writes selfing rate, mean fitness, mean nb of mutations per chromosome,
            // inbreeding depression and nb of fixed mutations in output file:
				
            fout << slf / twoN << " " << wbar << " " << le / twoN
				<< " " << 1 - (wself / wout) << " " << 1 - (wselfsel / woutsel) << " " << nbFix << endl;
        }
        
        for (i = 0; i < twoN; i++)
            pop[i] = temp[i];
    }
	
	fin = time(0);  
	
	fprintf(fichierS, "\n\nResults in file ");
	fprintf(fichierS, nomFichier);
	fprintf(fichierS, "\n");
	         
	// time length of simulation:
	int temps = difftime(fin, debut);
	fprintf(fichierS,
		 "\n%d generations took %d hour(s) %d minute(s) %d seconds\n",
		 gen, temps / 3600, (temps % 3600) / 60, temps % 60);
		
	// Date and time:
	ptr=localtime(&fin);
	fprintf(fichierS, asctime(ptr));
	
	delete [] pop;
	delete [] temp;
	delete [] Wij;
	delete [] Wijm;
}
